-------------------------------------------------------------------------------
--
-- Williams custom DMA IC SC1/SC2 (blitter)
--
-- (c) 2016-2021 Wolfgang Scherr, woz <at> pin4.at
-- all rights reserved, use at your own risk
-- http://www.pin4.at/ or https://blog.fh-kaernten.at/retrofusion/
--
-- Credits: http://seanriddle.com/blitter.html
--          http://seanriddle.com/blittest.html
--          (Thanks for your great web page, Sean!)
--
-------------------------------------------------------------------------------
-- Redistribution and use in source and synthesized forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
-- Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
--
-- Redistributions in synthesized form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
--
-- Neither the name of the author nor the names of other contributors may
-- be used to endorse or promote products derived from this software without
-- specific prior written permission.
--
-- The code and any derivatives must not be used for commercial purposes 
-- without specific prior written permission by the author.
--
-- THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.
--
-- Please report bugs to the author, but before you do so, please
-- make sure that this is not a derivative work and that
-- you have the latest version of this file.
-- $Id: SCx_blitter.vhd 1536 2021-01-05 20:47:20Z wolfi $
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity SCx_blitter is
	Generic ( -- for SC2 variant (newer Williams games), set this to false
		      sc_bug : boolean := true
			);
    Port ( clk_i : in  STD_LOGIC;   -- overall system clock 
           e_i : in  STD_LOGIC;	    -- edge sampling on rising E edge
           e_n_i : in  STD_LOGIC;	-- interface samples on falling E edge
           rstn_i : in  STD_LOGIC;  -- async. reset (low active)
           rd_wrn_i : in  STD_LOGIC;
           adr_i : in  STD_LOGIC_VECTOR (15 downto 0);
           cs_i : in  STD_LOGIC;
           data1_i : in  STD_LOGIC_VECTOR (7 downto 0);
           data2_i : in  STD_LOGIC_VECTOR (7 downto 0);
           ba_i : in  STD_LOGIC;
           bs_i : in  STD_LOGIC;
           halt_o : out  STD_LOGIC;
           wrun_o : out  STD_LOGIC;
           wrln_o : out  STD_LOGIC;
           adr_o : out  STD_LOGIC_VECTOR (15 downto 0);
           data_o : out  STD_LOGIC_VECTOR (7 downto 0)
         );
end SCx_blitter;

architecture RTL of SCx_blitter is

  signal reg_control_s : std_logic_vector(7 downto 0) := (others => '0');
  signal reg_mask_s    : std_logic_vector(7 downto 0) := (others => '0');
  signal reg_src_s     : std_logic_vector(15 downto 0) := (others => '0');
  signal reg_dst_s     : std_logic_vector(15 downto 0) := (others => '0');
  signal reg_w_s       : std_logic_vector(7 downto 0) := (others => '0');
  signal reg_h_s       : std_logic_vector(7 downto 0) := (others => '0');

  signal start_s        : std_logic := '0';
  signal stop_s         : std_logic := '1';

  type dma_state_t is (idle, rdbyte, wrbyte, rdslow, rdslow1);
  signal dma_state_s : dma_state_t := idle;

  signal buf_s       : std_logic_vector(11 downto 0) := (others => '0');

  signal cnt_src_s     : unsigned(15 downto 0) := (others => '0');
  signal cnt_dst_s     : unsigned(15 downto 0) := (others => '0');
  signal cnt_w_s       : unsigned(7 downto 0) := (others => '0');
  signal cnt_h_s       : unsigned(7 downto 0) := (others => '0');

begin
   -- Registers & DMA state machine, a reset stops it
  dmaproc : process (clk_i, rstn_i) is
  begin
    if rstn_i='0' then
      dma_state_s <= idle;
	  halt_o <= '0';
    elsif rising_edge(clk_i) then
	   -- register definition, it seems they are not affected by a reset,
	   -- values are kept after a finished blit and
 	   -- registers are write-only (according to seanriddle.com)
	   if e_n_i='1' then
		  if cs_i='1' and rd_wrn_i='0' then
			case to_integer(unsigned(adr_i(2 downto 0))) is
				when 0 => 
					reg_control_s <= data1_i;
					cnt_src_s <= unsigned(reg_src_s);
					cnt_dst_s <= unsigned(reg_dst_s);
					cnt_w_s <= unsigned(reg_w_s);
					cnt_h_s <= unsigned(reg_h_s);
					halt_o <= '1';
					buf_s <= (others => '0');
					if data1_i(2)='1' then
						dma_state_s <= rdslow;
					else
						dma_state_s <= rdbyte;
					end if;
				when 1 => reg_mask_s <= data1_i;
				when 2 => reg_src_s(15 downto 8) <= data1_i;
				when 3 => reg_src_s(7 downto 0) <= data1_i;
				when 4 => reg_dst_s(15 downto 8) <= data1_i;
				when 5 => reg_dst_s(7 downto 0) <= data1_i;
				when 6 => if sc_bug then reg_w_s <= data1_i xor "00000100";
									else reg_w_s <= data1_i; end if;
				when 7 => if sc_bug then reg_h_s <= data1_i xor "00000100";
									else reg_h_s <= data1_i; end if;
				when others => null;
			end case;
		  end if;
	   end if;
	   -- we need to act fast after writing the control register,
	   -- so we use the same process to avoid a delay of a clock and
	   -- keep a clean, synchronous setup based on both E edges
	   if e_i='1' or e_n_i='1' then
		  case dma_state_s is
			-- for slow RAM->RAM read, we can not read in E=0, so skip this phase
			when rdslow =>
			  -- we need to wait until the bus is free, we start in E=1 phase
			  if (ba_i and bs_i) = '1' then
				dma_state_s <= rdbyte;
			  end if;
			-- 1st DMA step, read from source (in E=1 phase when slow RAM->RAM)
			--               otherwise we read from the ROM card directly when E=0
			when rdbyte =>
			  -- we need to wait until the bus is free, we start in E=1 phase
			  if (ba_i and bs_i) = '1' then
				  if reg_control_s(5)='1' then
					-- shift right one nibble (pixel)
					buf_s <= buf_s(3 downto 0) & data2_i;
				  else
					-- just move
					buf_s <= data2_i & "0000";
				  end if;
				  if reg_control_s(2)='1' then
					dma_state_s <= rdslow1;
				  else
					dma_state_s <= wrbyte;
				  end if;
			  end if;
			-- for slow RAM->RAM read, we skip again E=0 before writing
			when rdslow1 =>
			  dma_state_s <= wrbyte;
			-- write in E=1 phase
			when wrbyte =>
			  if reg_control_s(2)='1' then
				dma_state_s <= rdslow;
			  else
				dma_state_s <= rdbyte;
			  end if;
			  -- reached W and H, done (--> stop)
			  if cnt_w_s<=1 and
				 cnt_h_s<=1 then
				halt_o <= '0';
			    dma_state_s <= idle;
			  -- reached W, reset positions
			  elsif cnt_w_s<=1 then
				cnt_h_s <= cnt_h_s-1;
				cnt_w_s <= unsigned(reg_w_s);
				cnt_src_s <= cnt_src_s + 1;
				if reg_control_s(0)='1' then -- in 256-stride, keep MSB
					cnt_src_s(15 downto 8) <= unsigned(reg_src_s(15 downto 8));
				end if;
				cnt_dst_s <= cnt_dst_s + 1;
				if reg_control_s(1)='1' then -- in 256-stride, keep MSB
					cnt_dst_s(15 downto 8) <= unsigned(reg_dst_s(15 downto 8));
				end if;
			  -- continue blitting, count positions
			  else
				cnt_w_s <= cnt_w_s-1; 
				if reg_control_s(0)='0' then
				  -- 1-stride
				  cnt_src_s <= cnt_src_s + 1;
				else
				  -- 256-stride
				  cnt_src_s <= cnt_src_s + 256;
				end if;
				if reg_control_s(1)='0' then
				  -- 1-stride
				  cnt_dst_s <= cnt_dst_s + 1;
				else
				  -- 256-stride
				  cnt_dst_s <= cnt_dst_s + 256;
				end if;
			  end if;
			when others => null;
		  end case;
       end if;
    end if;
  end process dmaproc;

  -- we pass simply the address bus from CPU through this circuit and let it handling the access
  adr_o <= std_logic_vector(cnt_src_s) when dma_state_s=rdbyte and (ba_i and bs_i)='1' else
        std_logic_vector(cnt_dst_s) when dma_state_s=wrbyte else
        adr_i;

  -- we pass simply the data bus from CPU through this circuit and let it handling the access
  data_o <= reg_mask_s      when dma_state_s=wrbyte and reg_control_s(4)='1' else -- solid mode
         buf_s(11 downto 4) when dma_state_s=wrbyte else
         data1_i;

   -- nibble handling of writes (even/odd pixels)
  wrun_o <= '0' when dma_state_s=wrbyte and reg_control_s(3)='1' and reg_control_s(7)='1' and buf_s(11 downto 8)="0000" else -- see http://seanriddle.com/blitter.html, block #14
            '1' when dma_state_s=wrbyte and reg_control_s(3)='1' and buf_s(11 downto 8)="0000" else -- foreground mode, don't copy
            reg_control_s(7) when dma_state_s=wrbyte else -- EVEN PIXEL (upper nibble) with write control
            rd_wrn_i;
  wrln_o <= '0' when dma_state_s=wrbyte and reg_control_s(3)='1' and reg_control_s(6)='1' and buf_s(7 downto 4)="0000" else -- see http://seanriddle.com/blitter.html, block #15
            '1' when dma_state_s=wrbyte and reg_control_s(3)='1' and buf_s(7 downto 4)="0000" else -- foreground mode, don't copy
            reg_control_s(6) when dma_state_s=wrbyte else -- ODD PIXEL (lower nibble) with write control
            rd_wrn_i;

end RTL;

