-------------------------------------------------------------------------------
--
-- Williams game board mapping to Digilent BASYS3 development platform
--
-- (c) 2016-2021 Wolfgang Scherr, woz <at> pin4.at
-- all rights reserved, use at your own risk
-- http://www.pin4.at/ or https://blog.fh-kaernten.at/retrofusion/
--
-------------------------------------------------------------------------------
-- Redistribution and use in source and synthesized forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
-- Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
--
-- Redistributions in synthesized form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
--
-- Neither the name of the author nor the names of other contributors may
-- be used to endorse or promote products derived from this software without
-- specific prior written permission.
--
-- The code and any derivatives must not be used for commercial purposes 
-- without specific prior written permission by the author.
--
-- THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.
--
-- Please report bugs to the author, but before you do so, please
-- make sure that this is not a derivative work and that
-- you have the latest version of this file.
-- $Id: basys3.vhd 1538 2021-01-07 15:00:39Z wolfi $
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library UNISIM;
use UNISIM.Vcomponents.all;

entity basys3 is
    Port ( -- 100MHz OSC
		   clk        : in  std_logic;
		   -- 16 switches
           sw         : in  std_logic_vector(15 downto 0);
		   -- 16 LEDs
           led        : out std_logic_vector(15 downto 0);
		   -- 7seg
           seg        : out std_logic_vector(6 downto 0);
           dp         : out std_logic;
           an         : out std_logic_vector(3 downto 0);
		   -- buttons
           btnC       : in  std_logic;
           btnU       : in  std_logic;
           btnL       : in  std_logic;
           btnR       : in  std_logic;
           btnD       : in  std_logic;
		   -- VGA
           vgaRed     : out std_logic_vector(3 downto 0);
           vgaBlue    : out std_logic_vector(3 downto 0);
           vgaGreen   : out std_logic_vector(3 downto 0);
           Hsync      : out std_logic;
           Vsync      : out std_logic
           );
end basys3;

-- BASYS 3 mappings:
--
--          switch_i						switch_s 	button_s
--
-- reset                                    15
--
-- WD off   2								12
-- blitter  1								14
-- stargate 0								13
--
--                     panelswitch_i
-- buttons: up     PB0 0						bU		0
--          down   PA7 1						bD		1
--          rev    PA6 2						bL		2
--          hyper  PA3 3					2
--          sbomb  PA2 4					3
--          thrust PA1 5						bR		3
--          fire   PA0 6						bC		4
--
-- 			1-P  PA5   7					0
--          2-P  PA4   8					8
--
--                          coinswitch_i
--           AUTO_UP        0				6
--           ADVANCE        1				7
--           HIGH SCORE RES 2				9
--           R COIN         3				10
--           L COIN         4				1
--           C COIN         5				11
--

architecture RTL of basys3 is

  --Inputs
	signal switch_s : std_logic_vector(15 downto 0);
	signal button_s : std_logic_vector(5 downto 0);

  --Video
	signal h : std_logic;
	signal v : std_logic;
	signal r : std_logic_vector(2 downto 0);
	signal g : std_logic_vector(2 downto 0);
	signal b : std_logic_vector(1 downto 0);

  --ROM
	signal oe   : std_logic;
	signal cs   : std_logic;
	signal adr  : std_logic_vector(15 downto 0);
	signal data : std_logic_vector(7 downto 0);
	signal datas12 : std_logic_vector(7 downto 0);
	signal datas11 : std_logic_vector(7 downto 0);
	signal datas10 : std_logic_vector(7 downto 0);
	signal datas9  : std_logic_vector(7 downto 0);
	signal datas8  : std_logic_vector(7 downto 0);
	signal datas7  : std_logic_vector(7 downto 0);
	signal datas6  : std_logic_vector(7 downto 0);
	signal datas5  : std_logic_vector(7 downto 0);
	signal datas4  : std_logic_vector(7 downto 0);
	signal datas3  : std_logic_vector(7 downto 0);
	signal datas2  : std_logic_vector(7 downto 0);
	signal datas1  : std_logic_vector(7 downto 0);

	-- Clock generation
	signal clk_fb       : std_logic;
	signal clk_fb_bufg  : std_logic;
	signal locked       : std_logic;
	signal clk_s        : std_logic;
	signal clk_en_s     : std_logic := '0';
	signal rst_s        : std_logic;

begin

	-----------------------------------------------------------------------
	-- Clock generation
	  u_a_mmcm : MMCME2_BASE
		generic map (
		  CLKFBOUT_MULT_F => 12.0, -- we want 48MHz  (--> 100*12/25)
		  CLKOUT0_DIVIDE_F => 25.0,	 -- we get 100MHz
		  CLKOUT0_DUTY_CYCLE => 0.5,
		  CLKFBOUT_PHASE => 0.0,
		  DIVCLK_DIVIDE => 1,
		  CLKIN1_PERIOD => 10.0
		)
		port map (
		  CLKFBOUT  => clk_fb,	-- must be 600-1600MHz	(100*12=1200MHz)
		  CLKOUT0   => clk_s,
		  CLKFBIN   => clk_fb_bufg,
		  CLKIN1    => clk,
		  --
		  PWRDWN    => '0',
		  LOCKED    => locked,
		  --
		  RST       => switch_s(15)
		);
	  u_dcm_a_clk_0_bufg   : BUFG port map (I => clk_fb,   O => clk_fb_bufg);

	  rst_s <= not locked;

	  clkdivin : process (clk_s) is
	  begin
		if rising_edge(clk_s) then
		   clk_en_s <= not clk_en_s; -- 24MHz enable for 48MHz
		end if;
	  end process clkdivin;

	-----------------------------------------------------------------------

	switch_s <= sw;
    button_s(0) <= btnU;
    button_s(1) <= btnD;
    button_s(2) <= btnL;
    button_s(3) <= btnR;
    button_s(4) <= btnC;

	-- MAIN board
   MAIN: ENTITY work.williams_main PORT MAP (
          clk_i => clk_s,
          clk_en_i => clk_en_s,
          rst_i => switch_s(15),

          switch_i => switch_s(14 downto 12),
          panelswitch_i(8) => switch_s(8),
          panelswitch_i(7) => switch_s(0),
          panelswitch_i(6 downto 5) => button_s(4 downto 3),
          panelswitch_i(4 downto 3) => switch_s(3 downto 2),
          panelswitch_i(2 downto 0) => switch_s(2 downto 0),
          coinswitch_i(5) => switch_s(11),
          coinswitch_i(4) => switch_s(1),
          coinswitch_i(3) => switch_s(10),
          coinswitch_i(2) => switch_s(9),
          coinswitch_i(1) => switch_s(7),
          coinswitch_i(0) => switch_s(6),

          led_o => led(3 downto 0),
          s7_o(6 downto 0) => seg,
          s7_o(7) => dp,
          s7sel_o => an,

		  r_o => r,
		  g_o => g,
		  b_o => b,
          hsync_o => h,
          vsync_o => v,

		  MemOE => oe,
          MemDB => data,
		  MemAdr => adr
        );

	led(15 downto 4) <= "000000000000";

	-----------------------------------------------------------------------
	-- Colour mapping (from 8bit to 12bit on BASYS3)

	Hsync <= not h;
	Vsync <= not v;
	vgaRed <= "1111" when r="111" else
			  "1101" when r="110" else
			  "1010" when r="101" else
			  "1000" when r="100" else
			  "0111" when r="011" else
			  "0101" when r="010" else
			  "0011" when r="001" else
			  "0000";
	vgaGreen <= "1111" when g="111" else
				"1101" when g="110" else
				"1010" when g="101" else
				"1000" when g="100" else
				"0111" when g="011" else
				"0101" when g="010" else
				"0011" when g="001" else
				"0000";
	vgaBlue <= "1101" when b="11" else
			   "1000" when b="10" else
			   "0101" when b="01" else
			   "0000";

	-----------------------------------------------------------------------
	-- ROM mapping to pre-initialised BRAMs

	cs <= not oe;

    i_rs1: ENTITY work.rs01 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas1
    );
    i_rs2: ENTITY work.rs02 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas2
    );
    i_rs3: ENTITY work.rs03 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas3
    );
    i_rs4: ENTITY work.rs04 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas4
    );
    i_rs5: ENTITY work.rs05 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas5
    );
    i_rs6: ENTITY work.rs06 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas6
    );
    i_rs7: ENTITY work.rs07 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas7
    );
    i_rs8: ENTITY work.rs08 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas8
    );
    i_rs9: ENTITY work.rs09 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas9
    );
    i_rs10: ENTITY work.rs10 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas10
    );
    i_rs11: ENTITY work.rs11 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas11
    );
    i_rs12: ENTITY work.rs12 PORT MAP (
        CLK => clk_s,
        ENA => cs,
        ADDR => adr(11 downto 0),
        DATA => datas12
    );
	------------------------------------------
	data <= -- these ROMs are always there, for every platform
			datas12 when adr(15 downto 12)="1111" else
		    datas11 when adr(15 downto 12)="1110" else
            datas10 when adr(15 downto 12)="1101" else
            -- 4k empty                     1100
            -- 4k empty                     1011
            -- 4k empty                     1010
            -- 4k empty                     1001
			-- these ROMS are either used in one continuous address space
			-- or mapped accordingly in pages of 4k to $C000 for Defender.
            datas9  when adr(15 downto 12)="1000" else
            datas8  when adr(15 downto 12)="0111" else
            datas7  when adr(15 downto 12)="0110" else
            datas6  when adr(15 downto 12)="0101" else
            datas5  when adr(15 downto 12)="0100" else
            datas4  when adr(15 downto 12)="0011" else
            datas3  when adr(15 downto 12)="0010" else
            datas2  when adr(15 downto 12)="0001" else
            datas1;  --                     0000

	-----------------------------------------------------------------------
	-- SOUND board (TODO)

	--SND: ENTITY work.williams_sound PORT MAP (
    --       clk => clk,
    --       clk_en => '1',
    --       rst => switch(13),
    --       diag => '0',
    --       dil => "00",
    --       sound_select => "0000000",
    --       sound_out => open --11 downto 0
    --     );

end architecture RTL;
