-------------------------------------------------------------------------------
--
-- Williams game board implementation (including I/O board and scan doubler)
--
-- (c) 2016-2021 Wolfgang Scherr, woz <at> pin4.at
-- all rights reserved, use at your own risk
-- http://www.pin4.at/ or https://blog.fh-kaernten.at/retrofusion/
--
-- This implementation is based on "defender theory" documents 
-- and many Williams schematics found on the internet.

-- Please note that newer schematics have similar circuitry, but renamed the
-- individual gates, so they may not match anymore with the notes here.
--
-------------------------------------------------------------------------------
-- Redistribution and use in source and synthesized forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
-- Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
--
-- Redistributions in synthesized form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
--
-- Neither the name of the author nor the names of other contributors may
-- be used to endorse or promote products derived from this software without
-- specific prior written permission.
--
-- The code and any derivatives must not be used for commercial purposes 
-- without specific prior written permission by the author.
--
-- THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.
--
-- Please report bugs to the author, but before you do so, please
-- make sure that this is not a derivative work and that
-- you have the latest version of this file.
-- $Id: williams_main.vhd 1540 2021-01-07 17:59:33Z wolfi $
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity williams_main is
    Port ( clk_i      : in  std_logic; -- Overall system clock
           clk_en_i   : in  std_logic; -- 24 MHz enable
           rst_i      : in  std_logic; -- high-active
		   -- VGA out
           r_o        : out std_logic_vector(2 downto 0);
           g_o        : out std_logic_vector(2 downto 0);
           b_o        : out std_logic_vector(1 downto 0);
           hsync_o    : out std_logic;
           vsync_o    : out std_logic;
		   -- ROM interface
           MemOE      : out std_logic;
           MemDB      : in  std_logic_vector(7 downto 0);	-- 8bit ROMs where a max. of
           MemAdr     : out std_logic_vector(15 downto 0);	-- $0000-$9FFF and $D000-$FFFF is used
		   -- Game control
           switch_i      : in  std_logic_vector(2 downto 0);
						-- switch:   STARGATE MODE  0
						--           BLITTER ON     1
						--           WATCHDOG OFF   2
           coinswitch_i  : in  std_logic_vector(5 downto 0);  -- (incomplete !!!!)
						-- coinsw:   AUTO_UP        0
						--           ADVANCE        1
						--           HIGH SCORE RES 2
						--           R COIN         3
						--           L COIN         4
						--           C COIN         5
           panelswitch_i : in  std_logic_vector(8 downto 0);
						-- buttons: up     PB0 0
						--          down   PA7 1
						--          rev    PA6 2
						--          hyper  PA3 3
						--          sbomb  PA2 4
						--          thrust PA1 5
						--          fire   PA0 6
						--
						-- 			1-P  PA5   7
						--          2-P  PA4   8
		   -- Debug stuff as these games offers
           s7sel_o       : out std_logic_vector(3 downto 0);  -- Stargate and newer
           s7_o          : out std_logic_vector(7 downto 0);  -- Stargate and newer
           led_o         : out std_logic_vector(3 downto 0)   -- Defender only
           );
end williams_main;

architecture RTL of williams_main is

-- Board variants selection
signal blitter  : boolean := true; 	-- blitter enabled (also stargate must be set true)
signal stargate : boolean := true;  -- different memory decoder ROMs / Flash range

-- Figure 3-2, Watchdog Circuit
signal wdr_ns       : std_logic := '1';
signal rst_s        : std_logic := '1';

-- Figure 3-3, MPU
signal mpu_aint_s   : std_logic_vector(15 downto 0);
signal mpu_a_s      : std_logic_vector(15 downto 0);
signal mpu_r_wn_s   : std_logic;
signal mpu_vma      : std_logic;
signal mpu_do_s     : std_logic_vector(7 downto 0);
signal mpu_doint_s  : std_logic_vector(7 downto 0);
signal mpu_di_s     : std_logic_vector(7 downto 0) := (others => '0');

signal mpu_lic_s    : std_logic;
signal mpu_ba_s     : std_logic;
signal mpu_bs_s     : std_logic;

signal mpu_halt_s   : std_logic;
signal mpu_halt_ns  : std_logic;

signal mpu_res_s    : std_logic;
signal mpu_res_ns   : std_logic;

signal blit_sel_s    : std_logic;

signal blit_h_r_wn_s   : std_logic;
signal blit_l_r_wn_s   : std_logic;

-- Figure 3-4, Clock Generator
signal clk_12meg_ns : std_logic;
signal clk_12meg_s  : std_logic;
signal clk_6meg_s   : std_logic;
signal clk_6meg_ns  : std_logic;
signal clk_4meg_s   : std_logic;
signal clk_4meg_ns  : std_logic;

-- Figure 3-5, E and Q Generator
signal clk_e_s      : std_logic;
signal clk_q_s      : std_logic;
signal clk_q_n_s    : std_logic;
signal clk_e_n_s    : std_logic;
signal q_s          : std_logic;
signal e_s          : std_logic;

-- Figure 3-1, Reset Generator
signal cr_ns        : std_logic := '0';

-- Figure 3-6, Video Address Generator
signal h_s          : unsigned(7 downto 0);
signal v_s          : unsigned(7 downto 0);
signal watchdog_ns  : std_logic;
signal va_s         : std_logic_vector(13 downto 0);

-- Figure 3-7, Video RAM Control
signal pa_s         : std_logic_vector(13 downto 8);

signal e1_ns        : std_logic;
signal e2_ns        : std_logic;
signal e3_ns        : std_logic;
signal we1h_ns       : std_logic;
signal we2h_ns       : std_logic;
signal we3h_ns       : std_logic;
signal we1l_ns       : std_logic;
signal we2l_ns       : std_logic;
signal we3l_ns       : std_logic;

signal mux1_s       : std_logic;
signal w_rn_s       : std_logic;

signal vp_s         : std_logic_vector(13 downto 6);

-- Figure 3-8, Video RAM Timing
signal we_s         : std_logic;
signal srl_ns       : std_logic;
--signal mux0_s       : std_logic;
--signal ras_ns       : std_logic;
--signal cas_ns       : std_logic;
signal q_4s5_s		: std_logic;
signal q_4s9_s 		: std_logic;
signal q_5s5_s 		: std_logic;

-- Figure 3-9, Video RAM Address Multiplexer
signal ma_s         : std_logic_vector(13 downto 0);

-- Figure 3-10, Video RAM
signal ram1_s        : std_logic_vector(7 downto 0);
signal ram2_s        : std_logic_vector(7 downto 0);
signal ram3_s        : std_logic_vector(7 downto 0);

-- Figure 3-11, Video Shift Registers
signal serial0_s       : std_logic;
signal serial1_s       : std_logic;
signal serial2_s       : std_logic;
signal serial3_s       : std_logic;
signal ser0_s : std_logic_vector(5 downto 0);
signal ser1_s : std_logic_vector(5 downto 0);
signal ser2_s : std_logic_vector(5 downto 0);
signal ser3_s : std_logic_vector(5 downto 0);

-- Figure 3-12, Color RAM Circuit
signal cram_sel_s     : std_logic;
signal r_s        : std_logic_vector(2 downto 0);
signal g_s        : std_logic_vector(2 downto 0);
signal b_s        : std_logic_vector(1 downto 0);

-- Figure 3-13, Monitor Sync Generator
signal hsync_s    : std_logic;
signal vsync_s    : std_logic;

-- Figure 3-14, Blanking Circuit
signal blank_s     : std_logic;
signal wdff_s      : std_logic;

-- Figure 3-15, Page 0 Decoder
signal ipia_cs_s        : std_logic;

-- Figure 3-16, CMOS RAM
signal cmos_sel_s : std_logic;
signal cmos_s     : std_logic_vector(7 downto 0);
signal cmos_adr_s : std_logic_vector(9 downto 0);
type cmos_array_t is array (integer range <>) of std_logic_vector(7 downto 0);
signal cmos : cmos_array_t(0 to 1023);

-- Figure 3-17, Vertical Count Buffer
signal vcnt_sel_s : std_logic;
signal vcnt_s     : std_logic_vector(7 downto 0);

-- Figure 3-18, Count 240
signal count_240_s : std_logic;
signal irq_4m_s    : std_logic;

-- Figure 3-19, ROM Board
signal rom_s     : std_logic_vector(7 downto 0);
signal rom_cs_ns     : std_logic;

-- Figure 3-20, Section/Page Select Decoder

signal psel_sel_s     : std_logic;
signal ps_s      : std_logic_vector(3 downto 0);
signal ps0_ns     : std_logic;
signal ps1_ns     : std_logic;
signal ps2_ns     : std_logic;
signal ps3_ns     : std_logic;
--signal ps4_ns     : std_logic;
--signal ps5_ns     : std_logic;
--signal ps6_ns     : std_logic;
signal ps7_ns     : std_logic;
--signal ps8_ns     : std_logic;
--signal ps9_ns     : std_logic;

-- Figure 3-23, ROM Board PIA
signal rpiairqa_ns    : std_logic;
signal rpiairqb_ns    : std_logic;
signal mpu_irq_ns     : std_logic;
signal soundsel_s     : std_logic_vector(5 downto 0);
signal pa_o_s         : std_logic_vector(5 downto 0);
signal pa_i_s         : std_logic_vector(5 downto 0);
signal rpia_cs_s      : std_logic;
signal rpia_do_s      : std_logic_vector(7 downto 0);
signal led_s     : std_logic_vector(3 downto 0);
signal x2        : std_logic := '0';
signal x1        : std_logic := '1';
signal low       : std_logic := '0';
signal high      : std_logic := '1';

-- Figure 3-24, Interface Board PIA
signal ipia_do_s     : std_logic_vector(7 downto 0);
signal ipia_ctrl1_s  : std_logic_vector(7 downto 0) := "00000000";
signal ipia_ctrl2_s  : std_logic_vector(7 downto 0) := "00000000";

-- VERILOG module embedded in VHDL
component mc6809e
port (
  CLK      : in  std_logic;
  E_N      : in  std_logic;
  Q_N      : in  std_logic;
  D        : in  std_logic_vector(7 downto 0);
  DOut     : out std_logic_vector(7 downto 0);
  ADDR     : out std_logic_vector(15 downto 0);
  RnW      : out std_logic;
  BS       : out std_logic;
  BA       : out std_logic;
  nIRQ     : in  std_logic;
  nFIRQ    : in  std_logic;
  nNMI     : in  std_logic;
  AVMA     : out std_logic;
  BUSY     : out std_logic;
  LIC      : out std_logic;
  nHALT    : in  std_logic;
  nRESET   : in  std_logic
);
end component;

begin

-- Multi-game config stuff
--

stargate <= switch_i(0)='1';
blitter <= switch_i(1)='1';

-- Figure 3-1, Reset Generator
--

rstgen : process (clk_i, rst_i) is
begin
  if rst_i='1' then
    rst_s <= '1';
  elsif rising_edge(clk_i) then
    if clk_en_i='1' then
      rst_s <= '0';
    end if;
    if rst_s='1' then
      cr_ns <= '0';
    elsif e_s='1' then -- note exactly the same, but will work out
      cr_ns <= '1';
    end if;
  end if;
end process rstgen;

-- Figure 3-2, Watchdog Circuit
--

watchdog : process (watchdog_ns, mpu_a_s, mpu_do_s, wdff_s) is
variable wdcnt : unsigned (3 downto 0);
begin
   if wdff_s='1' and mpu_a_s(9 downto 2)="11111111" and mpu_do_s(7 downto 1)="0011100" then
      wdcnt:="0000";
      wdr_ns <= '1';
   elsif rising_edge(watchdog_ns) then
      wdr_ns <= '1';
      wdcnt := wdcnt + 1;
      if wdcnt(3)='1' and switch_i(2)='0' then -- we can disable the WD
        wdr_ns <= '0';
      end if;
   end if;
end process watchdog;

-- Figure 3-4, Clock Generator
-- Figure 3-5, E and Q Generator
--
-- Breaking down the clock enables for a
-- sync. design, we ideally use 24MHz (and up)
-- as clock base to generate them like this:
--
--  12m (12Meg)  #.#.#.#.#.#. #.#.#.#.#.#. #.#.#.#.#.#.
-- f12m (!12Meg) .#.#.#.#.#.# .#.#.#.#.#.# .#.#.#.#.#.#
--   6m (6Meg)   #...#...#... #...#...#... #...#...#...
--  f6m (!6Meg)  ..#...#...#. ..#...#...#. ..#...#...#.
--   4m (4Meg)   #.....#..... #.....#..... #.....#.....
--  f4m (!4Meg)  ...#.....#.. ...#.....#.. ...#.....#..
--                           :            :
--  24m          |||||||||||| |||||||||||| ||||||||||||
-- clkdiv        0123456789AB 0123456789AB 0123456789AB
-- eq_cyc        000000111111 222222333333 000000111111
--                           :            :
--    e          .........#.. ............ .........#..
--    q          ...#........ ............ ...#........
--                           :            :
--    e_n        ............ .........#.. ............
--    q_n        ............ ...#........ ............
--                           :            :
--    e          ..........** **********.. ..........**
--    q          ....******** ****........ ....********
--                           :            :
--    Active RAM address bus cycles:
--                           :            :
--    pixelshift +  ++      + +  ++  ++  + +  ++      +
--    pixel-load        ++                        ++
--                           :            :
--    VIDEO      ##########__ __________## #########___
--    CPU        __________## ##########__ _________###
--                           :            :
--    Blitter    --rd.ROM--~~ rd/wr RAM~-- -rd.ROM--~~~
--
--    --> Thus, blitter only works with stargate ROM paging
--
clkgen : process (clk_i) is
  variable clkdiv : unsigned(3 downto 0);
  variable eq_cyc : unsigned(1 downto 0);
begin
	if rising_edge(clk_i) then
		if rst_s='1' then
			clkdiv := (others => '0');
			eq_cyc := (others => '0');
			clk_12meg_ns <= '0';
			clk_12meg_s <= '0';
			clk_6meg_s <= '0';
			clk_6meg_ns <= '0';
			clk_4meg_s <= '0';
			clk_4meg_ns <= '0';
			clk_q_s <= '0';
			clk_q_n_s <= '0';
			clk_e_s <= '0';
			clk_e_n_s <= '0';
			e_s <= '0';
			q_s <= '0';
		else
			if clk_e_s='1' then
				e_s <= '1';
			elsif clk_e_n_s='1' then
				e_s <= '0';
			elsif clk_q_s='1' then
				q_s <= '1';
			elsif clk_q_n_s='1' then
				q_s <= '0';
			end if;
			clk_12meg_ns <= '0';
			clk_12meg_s <= '0';
			clk_6meg_s <= '0';
			clk_6meg_ns <= '0';
			clk_4meg_s <= '0';
			clk_4meg_ns <= '0';
			clk_q_n_s <= '0';
			clk_e_n_s <= '0';
			clk_e_s <= '0';
			clk_q_s <= '0';
			if clk_en_i='1' then
				clkdiv := clkdiv + 1;
				if clkdiv=12 then
					clkdiv:=(others => '0');
				end if;
				clk_12meg_ns <= clkdiv(0);
				clk_12meg_s <= not clkdiv(0);
				if clkdiv(1 downto 0)="00" then
					clk_6meg_ns <= cr_ns; -- on sync. reset, this is kept low
				end if;
				if clkdiv(1 downto 0)="10" then
					clk_6meg_s <= cr_ns; -- on sync. reset, this is kept low
				end if;
				if clkdiv=0 or clkdiv=6 then
					clk_4meg_s <= '1';
					eq_cyc := eq_cyc + 1;
				elsif clkdiv=3 or clkdiv=9 then
					clk_4meg_ns <= '1';
					if eq_cyc=0 then
					  clk_q_s <= '1';
					elsif eq_cyc=1 then
					  clk_e_s <= '1';
					elsif eq_cyc=2 then
					  clk_q_n_s <= '1';
					elsif eq_cyc=3 then
					  clk_e_n_s <= '1';
					end if;
				end if;
			end if;
		end if;
	end if;
end process clkgen;

-- Figure 3-3, MPU
--

mpu_res_s <= rst_s or not(wdr_ns);
mpu_res_ns <= not mpu_res_s;
mpu_halt_ns <= not mpu_halt_s;

MCU : mc6809e
port map (
      CLK => clk_i,
      E_N => clk_e_n_s,
      Q_N => clk_q_n_s,
      nRESET => mpu_res_ns,
      nNMI => '1',
      nIRQ => mpu_irq_ns,
      nFIRQ => '1',
      nHALT => mpu_halt_ns,
	  BA => mpu_ba_s,
      BS => mpu_bs_s,
      D => mpu_di_s,
      Dout => mpu_doint_s,
      ADDR => mpu_aint_s,
      RnW => mpu_r_wn_s,
	  LIC => mpu_lic_s,
	  AVMA => mpu_vma,
	  BUSY => open
);

mpu_di_s <= -- ROMs
            rom_s when rom_cs_ns='0' else -- has higher prio than RAM read if enabled
            -- RAMs
            ram1_s when e1_ns='0' else
            ram2_s when e2_ns='0' else
            ram3_s when e3_ns='0' else
            -- SFRs
            cmos_s when cmos_sel_s='1' else
            vcnt_s when vcnt_sel_s='1' else
            rpia_do_s when rpia_cs_s='1' else
            ipia_do_s when ipia_cs_s='1' else
            -- DIRECT FEEDBACK
            mpu_do_s;

-- DMA via SC1 (SC2) Blitter Unit
--
DMA : entity work.SCx_blitter
port map (
      clk_i => clk_i,
      rstn_i => mpu_res_ns,
      e_i => clk_e_s,		-- instead of clk_4m + e, we directly use both e edges
      e_n_i => clk_e_n_s,
      --
		rd_wrn_i => mpu_r_wn_s,
		adr_i => mpu_aint_s,    -- address from MPU
		cs_i => blit_sel_s,
		data1_i => mpu_doint_s, -- data from MPU
		data2_i => mpu_di_s,    -- data from memory
		ba_i => mpu_ba_s,
		bs_i => mpu_bs_s,
      --
		halt_o => mpu_halt_s,
		wrun_o => blit_h_r_wn_s,
		wrln_o => blit_l_r_wn_s,
		adr_o => mpu_a_s,     -- address to memory
		data_o => mpu_do_s    -- data to memory
);

-- we can use the blitter only with the new game boards (after defender and stargate)
blit_sel_s <= '1' when mpu_a_s(15 downto 8)=x"CA" and blitter and stargate else
              '0';

-- Figure 3-6, Video Address Generator
--

vagen : process (clk_i, h_s, v_s) is
  variable q_4e5   : std_logic;
  variable pe_5e15 : std_logic;
  variable pe_5j15 : std_logic;
begin
  if h_s=x"FF" and v_s=x"FF" then
    pe_5e15:='1';
  else
    pe_5e15:='0';
  end if;
  if h_s(3 downto 0)=x"F" then
    pe_5j15:='1';
  else
    pe_5j15:='0';
  end if;
  if rising_edge(clk_i) then
    if cr_ns='0' then
      h_s <= (others => '0');
      v_s <= (others => '0');
      q_4e5 := '0';
    elsif clk_4meg_s='1' then
      h_s <= h_s + 1;
      if h_s=x"FF" then -- increment V
        if watchdog_ns='0' then
          v_s <= x"FC";
        else
          v_s <= v_s + 1;
           if v_s = x"FE" then
            q_4e5 := not q_4e5;
          elsif v_s(7)='0' then
            q_4e5 := '0';
          end if;
        end if;
      end if;
    end if;
    watchdog_ns <= not( q_4e5 and pe_5j15 and pe_5e15 );
  end if;
end process vagen;

va_s <= std_logic_vector(v_s) & std_logic_vector(h_s(7 downto 2));

-- Figure 3-7, Video RAM Control
--

w_rn_s <= not mpu_r_wn_s;

ramctrl : process (mpu_a_s,mpu_r_wn_s,we_s,e_s,va_s,stargate,blit_h_r_wn_s,blit_l_r_wn_s,srl_ns,clk_4meg_s) is
  variable lut_replace : unsigned(7 downto 0);
  variable lut : std_logic_vector(7 downto 0);
  type lut_array_t is array (integer range <>) of std_logic_vector(7 downto 0);
  -- we ignore upside-down mode, thus only 8x8 bit ROMs
  constant lut_array_decoder1 : lut_array_t(0 to 255) := (
   x"00",x"40",x"80",x"01",x"41",x"81",x"02",x"42",x"82",x"03",x"43",x"83",x"04",x"44",x"84",x"05",
   x"45",x"85",x"06",x"46",x"86",x"07",x"47",x"87",x"08",x"48",x"88",x"09",x"49",x"89",x"0A",x"4A",
   x"8A",x"0B",x"4B",x"8B",x"0C",x"4C",x"8C",x"0D",x"4D",x"8D",x"0E",x"4E",x"8E",x"0F",x"4F",x"8F",
   x"10",x"50",x"90",x"11",x"51",x"91",x"12",x"52",x"92",x"13",x"53",x"93",x"14",x"54",x"94",x"15",
   x"55",x"95",x"16",x"56",x"96",x"17",x"57",x"97",x"18",x"58",x"98",x"19",x"59",x"99",x"1A",x"5A",
   x"9A",x"1B",x"5B",x"9B",x"1C",x"5C",x"9C",x"1D",x"5D",x"9D",x"1E",x"5E",x"9E",x"1F",x"5F",x"9F",
   x"20",x"60",x"A0",x"21",x"61",x"A1",x"22",x"62",x"A2",x"23",x"63",x"A3",x"24",x"64",x"A4",x"25",
   x"65",x"A5",x"26",x"66",x"A6",x"27",x"67",x"A7",x"28",x"68",x"A8",x"29",x"69",x"A9",x"2A",x"6A",
   x"AA",x"2B",x"6B",x"AB",x"2C",x"6C",x"AC",x"2D",x"6D",x"AD",x"2E",x"6E",x"AE",x"2F",x"6F",x"AF",
   x"30",x"70",x"B0",x"31",x"71",x"B1",x"32",x"72",x"B2",x"33",x"73",x"B3",x"34",x"74",x"B4",x"35",
   x"75",x"B5",x"36",x"76",x"B6",x"37",x"77",x"B7",x"38",x"78",x"B8",x"39",x"79",x"B9",x"3A",x"7A",
   x"BA",x"3B",x"7B",x"BB",x"3C",x"7C",x"BC",x"3D",x"7D",x"BD",x"3E",x"7E",x"BE",x"3F",x"7F",x"BF",
   x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",
   x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",
   x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",
   x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00",x"00"
  );
  constant lut_array_decoder4 : lut_array_t(0 to 255) := (
   x"81",x"02",x"42",x"82",x"03",x"43",x"83",x"04",x"44",x"84",x"05",x"45",x"85",x"06",x"46",x"86",
   x"07",x"47",x"87",x"08",x"48",x"88",x"09",x"49",x"89",x"0A",x"4A",x"8A",x"0B",x"4B",x"8B",x"0C",
   x"4C",x"8C",x"0D",x"4D",x"8D",x"0E",x"4E",x"8E",x"0F",x"4F",x"8F",x"10",x"50",x"90",x"11",x"51",
   x"91",x"12",x"52",x"92",x"13",x"53",x"93",x"14",x"54",x"94",x"15",x"55",x"95",x"16",x"56",x"96",
   x"17",x"57",x"97",x"18",x"58",x"98",x"19",x"59",x"99",x"1A",x"5A",x"9A",x"1B",x"5B",x"9B",x"1C",
   x"5C",x"9C",x"1D",x"5D",x"9D",x"1E",x"5E",x"9E",x"1F",x"5F",x"9F",x"20",x"60",x"A0",x"21",x"61",
   x"A1",x"22",x"62",x"A2",x"23",x"63",x"A3",x"24",x"64",x"A4",x"25",x"65",x"A5",x"26",x"66",x"A6",
   x"27",x"67",x"A7",x"28",x"68",x"A8",x"29",x"69",x"A9",x"2A",x"6A",x"AA",x"2B",x"6B",x"AB",x"2C",
   x"6C",x"AC",x"2D",x"6D",x"AD",x"2E",x"6E",x"AE",x"2F",x"6F",x"AF",x"30",x"70",x"B0",x"31",x"71",
   x"B1",x"32",x"72",x"B2",x"33",x"73",x"B3",x"34",x"74",x"B4",x"35",x"75",x"B5",x"36",x"76",x"B6",
   x"37",x"77",x"B7",x"38",x"78",x"B8",x"39",x"79",x"B9",x"3A",x"7A",x"BA",x"3B",x"7B",x"BB",x"3C",
   x"7C",x"BC",x"3D",x"7D",x"BD",x"3E",x"7E",x"BE",x"3F",x"7F",x"BF",x"00",x"40",x"80",x"01",x"41",
   x"C3",x"CF",x"D0",x"D9",x"D2",x"C9",x"C7",x"C8",x"D4",x"80",x"A8",x"C3",x"A9",x"80",x"B1",x"B9",
   x"B8",x"B1",x"80",x"D7",x"C9",x"CC",x"CC",x"C9",x"C1",x"CD",x"D3",x"80",x"C5",x"CC",x"C5",x"C3",
   x"D4",x"D2",x"CF",x"CE",x"C9",x"C3",x"D3",x"80",x"C9",x"CE",x"C3",x"AE",x"80",x"C1",x"CC",x"CC",
   x"80",x"D2",x"C9",x"C7",x"C8",x"D4",x"D3",x"80",x"D2",x"C5",x"D3",x"C5",x"D2",x"D6",x"C5",x"C4"
   );
  constant lut_array_decoder5 : lut_array_t(0 to 255) := (
   x"7F",x"00",x"01",x"02",x"03",x"04",x"05",x"06",x"07",x"08",x"09",x"0A",x"0B",x"0C",x"0D",x"0E",
   x"0F",x"10",x"11",x"12",x"13",x"14",x"15",x"16",x"17",x"18",x"19",x"1A",x"1B",x"1C",x"1D",x"1E",
   x"1F",x"20",x"21",x"22",x"23",x"24",x"25",x"26",x"27",x"28",x"29",x"2A",x"2B",x"2C",x"2D",x"2E",
   x"2F",x"30",x"31",x"32",x"33",x"34",x"35",x"36",x"37",x"38",x"39",x"3A",x"3B",x"3C",x"3D",x"3E",
   x"3F",x"40",x"41",x"42",x"43",x"44",x"45",x"46",x"47",x"48",x"49",x"4A",x"4B",x"4C",x"4D",x"4E",
   x"4F",x"50",x"51",x"52",x"53",x"54",x"55",x"56",x"57",x"58",x"59",x"5A",x"5B",x"5C",x"5D",x"5E",
   x"5F",x"60",x"61",x"62",x"63",x"64",x"65",x"66",x"67",x"68",x"69",x"6A",x"6B",x"6C",x"6D",x"6E",
   x"6F",x"70",x"71",x"72",x"73",x"74",x"75",x"76",x"77",x"78",x"79",x"7A",x"7B",x"7C",x"7D",x"7E",
   x"7F",x"80",x"81",x"82",x"83",x"84",x"85",x"86",x"87",x"88",x"89",x"8A",x"8B",x"8C",x"8D",x"8E",
   x"8F",x"90",x"91",x"92",x"93",x"94",x"95",x"96",x"97",x"98",x"99",x"9A",x"9B",x"9C",x"9D",x"9E",
   x"9F",x"A0",x"A1",x"A2",x"A3",x"A4",x"A5",x"A6",x"A7",x"A8",x"A9",x"AA",x"AB",x"AC",x"AD",x"AE",
   x"AF",x"B0",x"B1",x"B2",x"B3",x"B4",x"B5",x"B6",x"B7",x"B8",x"B9",x"BA",x"BB",x"BC",x"BD",x"BE",
   x"BF",x"C0",x"C1",x"C2",x"C3",x"C4",x"C5",x"C6",x"C7",x"C8",x"C9",x"CA",x"CB",x"CC",x"CD",x"CE",
   x"CF",x"D0",x"D1",x"D2",x"D3",x"D4",x"D5",x"D6",x"D7",x"D8",x"D9",x"DA",x"DB",x"DC",x"DD",x"DE",
   x"DF",x"E0",x"E1",x"E2",x"E3",x"E4",x"E5",x"E6",x"E7",x"E8",x"E9",x"EA",x"EB",x"EC",x"ED",x"EE",
   x"EF",x"F0",x"F1",x"F2",x"F3",x"F4",x"F5",x"F6",x"F7",x"FF",x"FF",x"FF",x"FF",x"FF",x"FF",x"FF"
   );
	-- bugfix of decoder 5 in newer games
  constant lut_array_decoder6 : lut_array_t(0 to 255) := (
   x"00",x"00",x"01",x"02",x"03",x"04",x"05",x"06",x"07",x"08",x"09",x"0A",x"0B",x"0C",x"0D",x"0E",
   x"0F",x"10",x"11",x"12",x"13",x"14",x"15",x"16",x"17",x"18",x"19",x"1A",x"1B",x"1C",x"1D",x"1E",
   x"1F",x"20",x"21",x"22",x"23",x"24",x"25",x"26",x"27",x"28",x"29",x"2A",x"2B",x"2C",x"2D",x"2E",
   x"2F",x"30",x"31",x"32",x"33",x"34",x"35",x"36",x"37",x"38",x"39",x"3A",x"3B",x"3C",x"3D",x"3E",
   x"3F",x"40",x"41",x"42",x"43",x"44",x"45",x"46",x"47",x"48",x"49",x"4A",x"4B",x"4C",x"4D",x"4E",
   x"4F",x"50",x"51",x"52",x"53",x"54",x"55",x"56",x"57",x"58",x"59",x"5A",x"5B",x"5C",x"5D",x"5E",
   x"5F",x"60",x"61",x"62",x"63",x"64",x"65",x"66",x"67",x"68",x"69",x"6A",x"6B",x"6C",x"6D",x"6E",
   x"6F",x"70",x"71",x"72",x"73",x"74",x"75",x"76",x"77",x"78",x"79",x"7A",x"7B",x"7C",x"7D",x"7E",
   x"7F",x"80",x"81",x"82",x"83",x"84",x"85",x"86",x"87",x"88",x"89",x"8A",x"8B",x"8C",x"8D",x"8E",
   x"8F",x"90",x"91",x"92",x"93",x"94",x"95",x"96",x"97",x"98",x"99",x"9A",x"9B",x"9C",x"9D",x"9E",
   x"9F",x"A0",x"A1",x"A2",x"A3",x"A4",x"A5",x"A6",x"A7",x"A8",x"A9",x"AA",x"AB",x"AC",x"AD",x"AE",
   x"AF",x"B0",x"B1",x"B2",x"B3",x"B4",x"B5",x"B6",x"B7",x"B8",x"B9",x"BA",x"BB",x"BC",x"BD",x"BE",
   x"BF",x"C0",x"C1",x"C2",x"C3",x"C4",x"C5",x"C6",x"C7",x"C8",x"C9",x"CA",x"CB",x"CC",x"CD",x"CE",
   x"CF",x"D0",x"D1",x"D2",x"D3",x"D4",x"D5",x"D6",x"D7",x"D8",x"D9",x"DA",x"DB",x"DC",x"DD",x"DE",
   x"DF",x"E0",x"E1",x"E2",x"E3",x"E4",x"E5",x"E6",x"E7",x"E8",x"E9",x"EA",x"EB",x"EC",x"ED",x"EE",
   x"EF",x"F0",x"F1",x"F2",x"F3",x"F4",x"F5",x"F6",x"F7",x"00",x"00",x"00",x"00",x"00",x"00",x"00"
   );

begin
  if stargate then
    lut := lut_array_decoder4(to_integer(unsigned(mpu_a_s(15 downto 8))));
    vp_s(13 downto 6) <= lut_array_decoder6(to_integer(unsigned(va_s(13 downto 6))));
  else
    lut := lut_array_decoder1(to_integer(unsigned(mpu_a_s(15 downto 8))));   
    vp_s(13 downto 6) <= va_s(13 downto 6); -- direct mapping (when using decoder 1 only)
  end if;
  pa_s(13 downto 8) <= lut(5 downto 0);

  -- for stargate, we control the read by the priotity of the ROM data multiplexed to the CPU input
  -- so we handle this access the same way as defender (all below $C000 enables RAM)
  e1_ns <= '1';
  e2_ns <= '1';
  e3_ns <= '1';
  if (lut(7 downto 6)="00" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and mpu_r_wn_s='1') then
    e1_ns <= '0';
  end if;
  if (lut(7 downto 6)="01" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and mpu_r_wn_s='1') then
    e2_ns <= '0';
  end if;
  if (lut(7 downto 6)="10" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and mpu_r_wn_s='1') then
    e3_ns <= '0';
  end if;

  -- write is always to RAM in defender and stargate
  we1h_ns <= '1';
  we2h_ns <= '1';
  we3h_ns <= '1';
  mux1_s <= '1';
  if lut(7 downto 6)="00" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_h_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we1h_ns <= '0';
  end if;
  if lut(7 downto 6)="01" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_h_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we2h_ns <= '0';
  end if;
  if lut(7 downto 6)="10" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_h_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we3h_ns <= '0';
  end if;
  if (mpu_a_s(15)='0' or mpu_a_s(14)='0') and e_s='1' then
    mux1_s <= '0';
  end if;
  we1l_ns <= '1';
  we2l_ns <= '1';
  we3l_ns <= '1';
  mux1_s <= '1';
  if lut(7 downto 6)="00" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_l_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we1l_ns <= '0';
  end if;
  if lut(7 downto 6)="01" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_l_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we2l_ns <= '0';
  end if;
  if lut(7 downto 6)="10" and (mpu_a_s(15)='0' or mpu_a_s(14)='0') and blit_l_r_wn_s='0' and we_s='1' and e_s='1' and clk_4meg_s='1' then
    we3l_ns <= '0';
  end if;
  if (mpu_a_s(15)='0' or mpu_a_s(14)='0') and e_s='1' then
    mux1_s <= '0';
  end if;
end process ramctrl;

-- Figure 3-8, Video RAM Timing
--

ramtim : process (clk_i) is
begin
  if rising_edge(clk_i) then
    if rst_s='1' then
      q_4s5_s <= '0';
      q_4s9_s <= '0';
      q_5s5_s <= '0';
    elsif clk_12meg_ns='1' then
      q_4s5_s <= e_s xor q_s;
      q_4s9_s <= not q_4s5_s;
    elsif clk_12meg_s='1' then
      q_5s5_s <= not q_4s9_s;
    end if;
-- We don't need that:
--    ras_ns <= q_4s5 and not q_4s9;
--    cas_ns <= q_5s5;
--    mux0_s <= q_4s9;
  end if;
end process ramtim;

we_s <= not q_5s5_s;
srl_ns <= not(q_s and (q_4s5_s and not q_4s9_s) and we_s);

-- Figure 3-9, Video RAM Address Multiplexer
--

ma_s <= mpu_a_s(7 downto 0) & pa_s(13 downto 8) when mux1_s='0' else
        vp_s(13 downto 6) & va_s(5 downto 0);

-- Figure 3-10, Video RAM
--

ram1h : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(7 downto 4),
      data_o => ram1_s(7 downto 4),
      we_n  => we1h_ns
);
ram2h : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(7 downto 4),
      data_o => ram2_s(7 downto 4),
      we_n  => we2h_ns
);
ram3h : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(7 downto 4),
      data_o => ram3_s(7 downto 4),
      we_n  => we3h_ns
);
ram1l : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(3 downto 0),
      data_o => ram1_s(3 downto 0),
      we_n  => we1l_ns
);
ram2l : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(3 downto 0),
      data_o => ram2_s(3 downto 0),
      we_n  => we2l_ns
);
ram3l : entity work.ram4_16k 
port map (
      clk  => clk_i,
      address => ma_s(13 downto 0),
      data_i => mpu_do_s(3 downto 0),
      data_o => ram3_s(3 downto 0),
      we_n  => we3l_ns
);

-- Figure 3-11, Video Shift Registers
--

vshift : process (clk_i) is
  variable doshift : std_logic;
begin
  if rising_edge(clk_i) then
    -- select shifting/loading
    if srl_ns='0' then
      doshift:='0';
    elsif clk_4meg_ns='1' then
      doshift:='1';
    end if;
    -- shift or load data
    if clk_6meg_ns='1' then   
      if doshift='1' then
         ser0_s <= '0' & ser0_s(5 downto 1);
         ser1_s <= '0' & ser1_s(5 downto 1);
         ser2_s <= '0' & ser2_s(5 downto 1);
         ser3_s <= '0' & ser3_s(5 downto 1);
      else
         ser0_s <= ram3_s(0)&ram3_s(4)&ram2_s(0)&ram2_s(4)&ram1_s(0)&ram1_s(4);
         ser1_s <= ram3_s(1)&ram3_s(5)&ram2_s(1)&ram2_s(5)&ram1_s(1)&ram1_s(5);
         ser2_s <= ram3_s(2)&ram3_s(6)&ram2_s(2)&ram2_s(6)&ram1_s(2)&ram1_s(6);
         ser3_s <= ram3_s(3)&ram3_s(7)&ram2_s(3)&ram2_s(7)&ram1_s(3)&ram1_s(7);
      end if;
    end if;
  end if;
end process vshift;

serial0_s <= ser0_s(0);
serial1_s <= ser1_s(0);
serial2_s <= ser2_s(0);
serial3_s <= ser3_s(0);

-- Figure 3-12, Color RAM Circuit
--

cram_sel_s <= '1' when clk_q_s='1' and mpu_a_s(11 downto 10)="00" and mpu_a_s(15 downto 12)=x"C" and w_rn_s='1' and ps0_ns='0' and mpu_a_s(4)='0' and not(stargate) else
              '1' when clk_q_s='1' and w_rn_s='1' and mpu_a_s(11 downto 10)="00" and mpu_a_s(15 downto 12)=x"C" and stargate else 
              '0';

cram : process (clk_i,serial0_s,serial1_s,serial2_s,serial3_s,blank_s) is
  type cram_array_t is array (integer range <>) of std_logic_vector(7 downto 0);
  variable cram : cram_array_t(0 to 15);
  variable adr : std_logic_vector(3 downto 0);  
  variable rgb : std_logic_vector(7 downto 0);  
begin
  if rising_edge(clk_i) then
    if cram_sel_s='1' then
      cram(to_integer(unsigned(mpu_a_s(3 downto 0)))) := mpu_do_s;
    end if;
  end if;
  adr := serial3_s & serial2_s & serial1_s & serial0_s;
  rgb := cram(to_integer(unsigned(adr)));
  if blank_s='1' then
    b_s <= "00";
    g_s <= "000";
    r_s <= "000";
  else
    b_s <= rgb(7 downto 6);
    g_s <= rgb(5 downto 3);
    r_s <= rgb(2 downto 0);
  end if;
end process cram;

wdff_s <= '1' when mpu_a_s(11 downto 10)="00" and mpu_a_s(15 downto 12)=x"C" and w_rn_s='1' and ps0_ns='0' and mpu_a_s(4)='1' and not(stargate) else
          '1' when mpu_a_s(15 downto 8)=x"CB" and stargate else 
          '0';

-- Figure 3-13, Monitor Sync Generator
--

vsync_s <= '1' when va_s(13 downto 9)="11111" else '0';
hsync_s <= '1' when va_s(5 downto 2)="1110" else '0';

-- Figure 3-14, Blanking Circuit
--

blanking : process (clk_i) is -- (va_s, blank_s) is
begin
  if rising_edge(clk_i) then -- we may replace the RS latch by something synchronous....
	  if va_s(5 downto 4)="11" and va_s(2)='1' then -- and va_s(0)='1' then   -- is too late....
		blank_s <= '1';
	  elsif va_s(1 downto 0)="11" and va_s(5)='0' and blank_s='1' then
		blank_s <= '0';
	  end if;
  end if;
end process blanking;

-- Figure 3-15, Page 0 Decoder
--

ipia_cs_s <= '1' when mpu_a_s(11 downto 10)="11" and mpu_a_s(4 downto 2)="001" and mpu_a_s(15 downto 12)=x"C" and ps0_ns='0' and not(stargate) else
             '1' when mpu_a_s(15 downto 8)=x"C8" and mpu_a_s(3 downto 2)="01" and stargate else
             '0';

rpia_cs_s <= '1' when mpu_a_s(11 downto 10)="11" and mpu_a_s(4 downto 2)="000" and mpu_a_s(15 downto 12)=x"C" and ps0_ns='0' and not(stargate) else
             '1' when mpu_a_s(15 downto 8)=x"C8" and mpu_a_s(3 downto 2)="11" and stargate else
             '0';

-- Figure 3-16, CMOS RAM
--

cmos_sel_s <= '1' when mpu_a_s(11 downto 10)="01" and mpu_a_s(15 downto 12)=x"C" and ps0_ns='0' and not(stargate) else
              '1' when mpu_a_s(11 downto 10)="11" and mpu_a_s(15 downto 12)=x"C" and stargate else
              '0';

cmos_adr_s <= "00" & mpu_a_s(7 downto 0) when not(stargate) else -- "ghost images" for smaller CMOS ram
              mpu_a_s(9 downto 0);

-- most games use only 4 bit CMOS-RAM, we implement the full 8bit width (which is nevertheless ignored if not used)
cmosproc : process (clk_i) is
begin
  if rising_edge(clk_i) then
    if cmos_sel_s='1' and clk_q_s='1' and mpu_r_wn_s='0' then
		cmos(to_integer(unsigned(cmos_adr_s))) <= mpu_do_s;
    end if;
 	 cmos_s <= cmos(to_integer(unsigned(cmos_adr_s)));
  end if;
end process cmosproc;

-- Figure 3-17, Vertical Count Buffer
--

vcnt_sel_s <= '1' when mpu_a_s(11 downto 10)="10" and mpu_a_s(15 downto 12)=x"C" and ps0_ns='0' and not(stargate) else 
              '1' when mpu_a_s(15 downto 8)=x"CB" and stargate else
				  '0';
vcnt_s <= va_s(13 downto 8) & "11";

-- Figure 3-18, Count 240
--

count_240_s <= '1' when va_s(13 downto 10)=x"F" else '0';
irq_4m_s <= va_s(11);

-- Figure 3-19, ROM Board
--

rom_cs_ns <= '0' when mpu_r_wn_s='1' and mpu_a_s(15 downto 14)="11" and (mpu_a_s(13) or mpu_a_s(12) or ps0_ns)='1' and not(stargate) else -- $C000 is paged in defender, only pg1 and up is ROM
             '0' when mpu_r_wn_s='1' and (unsigned(mpu_a_s(15 downto 12))>x"C" or (unsigned(mpu_a_s(15 downto 12))<x"9" and ps_s(0)='1')) and stargate else -- $0000-$9FFF is paged in stargate
			    '1';

MemOE <= rom_cs_ns;

MemAdr <= ps_s(3 downto 0) & mpu_a_s(11 downto 0) when mpu_a_s(15 downto 12)=x"C" and not(stargate) else -- paged 4k of ROM at $C000 + ROMs at $D000-$FFFF in Defender
          mpu_a_s(15 downto 0); -- single ROM/RAM page from $0000-$9FFF + ROMs at $D000-$FFFF for all other platforms

rom_s <= MemDB(7 downto 0);

-- Figure 3-20, Section/Page Select Decoder
--

psel_sel_s <= '1' when mpu_a_s(15 downto 8)=x"C9" and stargate else
              '1' when mpu_a_s(15 downto 12)=x"D" and mpu_a_s(11)='0' and not(stargate) else
				  '0';

pgsel : process (clk_i) is
begin
  if rising_edge(clk_i) then
    if rst_s='1' then
      ps_s <= (others => '0');
    elsif (clk_q_s='1' and mpu_r_wn_s='0' and psel_sel_s='1') then -- (rsel1)
      ps_s <= mpu_do_s(3 downto 0);
    end if;
  end if;
end process pgsel;   

ps0_ns <= '0' when ps_s=x"0" else '1';
ps1_ns <= '0' when ps_s=x"1" else '1';
ps2_ns <= '0' when ps_s=x"2" else '1';
ps3_ns <= '0' when ps_s=x"3" else '1';
--ps4_ns <= '0' when ps_s=x"4" else '1';
--ps5_ns <= '0' when ps_s=x"5" else '1';
--ps6_ns <= '0' when ps_s=x"6" else '1';
ps7_ns <= '0' when ps_s=x"7" else '1';
--ps8_ns <= '0' when ps_s=x"8" else '1';
--ps9_ns <= '0' when ps_s=x"9" else '1';

-- Figure 3-23, ROM Board PIA
--

-- INCOMPLETE !!!!!!

RPIA : entity work.mc6821
generic map ( instate_a => "00000000",
          instate_b => "11111111" )
port map (
      clk_i => clk_i,
      e_i => clk_e_s,
      e_n_i => clk_e_n_s,
      rstn_i => mpu_res_ns,
      rd_wrn_i => mpu_r_wn_s,
      cs0_i => rpia_cs_s,
	  cs1_i => '1',
	  cs2n_i => '0',

      rs_i => mpu_a_s(1 downto 0),
      data_i => mpu_do_s,
      data_o => rpia_do_s,

      irqan_o => rpiairqa_ns,
      irqbn_o => rpiairqb_ns,
      pa_o(7 downto 6) => led_s(3 downto 2),
	  pa_o(5 downto 0) => pa_o_s,
      pa_i(7 downto 6) => led_s(3 downto 2),
      pa_i(5 downto 0) => pa_i_s,
      ca1_i => count_240_s,
      ca2_i => x2,
      ca2_o => x2,
      pb_o(7 downto 6) => led_s(1 downto 0),
      pb_o(5 downto 0) => soundsel_s,
      pb_i(7 downto 6) => led_s(1 downto 0),
      pb_i(5 downto 0) => soundsel_s,
      cb1_i => irq_4m_s,
      cb2_i => x1,
      cb2_o => x1
);

-- it seems we need to do that ?
pa_i_s <= coinswitch_i or pa_o_s;

led_o <= led_s when not(stargate) else
		 "0000";

s7sel_o <= "1111" when not(stargate) else
           "1110";

s7_o <= "11111111" when not(stargate) else
        "11000000" when x2&x1&led_s(1 downto 0)=x"0" else
        "11111001" when x2&x1&led_s(1 downto 0)=x"1" else
        "10100100" when x2&x1&led_s(1 downto 0)=x"2" else
        "10000110" when x2&x1&led_s(1 downto 0)=x"3" else
        "10011001" when x2&x1&led_s(1 downto 0)=x"4" else
        "10010010" when x2&x1&led_s(1 downto 0)=x"5" else
        "10000010" when x2&x1&led_s(1 downto 0)=x"6" else
        "11111000" when x2&x1&led_s(1 downto 0)=x"7" else
        "10000000" when x2&x1&led_s(1 downto 0)=x"8" else
        "10010000" when x2&x1&led_s(1 downto 0)=x"9" else
		"01111111";

mpu_irq_ns <= rpiairqa_ns and rpiairqb_ns;

-- Figure 3-24, Interface Board Assembly
--

IPIA : entity work.mc6821
generic map ( instate_a => "00000000",
              instate_b => "00000000" )
port map (
      clk_i => clk_i,
      e_i => clk_e_s,
      e_n_i => clk_e_n_s,
      rstn_i => mpu_res_ns,
      rd_wrn_i => mpu_r_wn_s,
      cs0_i => ipia_cs_s,
	  cs1_i => '1',
	  cs2n_i => '0',

      rs_i => mpu_a_s(1 downto 0),
      data_i => mpu_do_s,
      data_o => ipia_do_s,

      irqan_o => open,
      irqbn_o => open,
      pa_o => open,
      pa_i => ipia_ctrl1_s,
      ca1_i => '1',
      ca2_i => '0',
      ca2_o => open,
      pb_o => open,
      pb_i => ipia_ctrl2_s,
      cb1_i => '1',
      cb2_i => '0',
      cb2_o => open
);

-- INCOMPLETE !!!!!!

usrctrl : process (panelswitch_i) is
begin
  ipia_ctrl1_s <= (others => '0');
  ipia_ctrl2_s <= (others => '0');
  if panelswitch_i(0)='1' then  -- up
	ipia_ctrl2_s(0) <= '1';
  end if;
  if panelswitch_i(1)='1' then  -- down
	ipia_ctrl1_s(7) <= '1';
  end if;
  if panelswitch_i(2)='1' then  -- rev
	ipia_ctrl1_s(6) <= '1';
  end if;
  if panelswitch_i(7)='1' then  -- PLAYER 1
	ipia_ctrl1_s(5) <= '1';
  end if;
  if panelswitch_i(8)='1' then  -- PLAYER 2
	ipia_ctrl1_s(4) <= '1';
  end if;
  if panelswitch_i(3)='1' then  -- hyperspace
	ipia_ctrl1_s(3) <= '1';
  end if;
  if panelswitch_i(4)='1' then  -- smartbomb
	ipia_ctrl1_s(2) <= '1';
  end if;
  if panelswitch_i(5)='1' then  -- thrust
	ipia_ctrl1_s(1) <= '1';
  end if;
  if panelswitch_i(6)='1' then  -- fire
	ipia_ctrl1_s(0) <= '1';
  end if;
end process usrctrl;

-------------------------------------------------------------------
-- convert NTSC timing to VGA timing by
-- doubling each video line generated
-------------------------------------------------------------------
scandoubler: block

	type scanline_t is array (0 to 383) of std_logic_vector(8 downto 0);
	signal scanline_s : scanline_t := (others => (others => '0'));

begin

	line_in : process (clk_i) is
	variable pixcnt_in : unsigned(8 downto 0) := (others => '0');
	variable pixcnt_out : unsigned(8 downto 0) := (others => '0');
	variable hsync : std_logic;
	begin
	  if rising_edge(clk_i) then
		-- start of sync resets line
		if hsync_s='1' and hsync='0' then
		  pixcnt_in := (others => '0');
		  pixcnt_out := (others => '0');
		else
		  if clk_6meg_ns='1' then
			 scanline_s(to_integer(pixcnt_in)) <= hsync_s & r_s & g_s & b_s;
			 pixcnt_in := pixcnt_in + 1;
		  end if;
		  if clk_12meg_ns='1' then
			 r_o <= scanline_s(to_integer(pixcnt_out))(7 downto 5);
			 g_o <= scanline_s(to_integer(pixcnt_out))(4 downto 2);
			 b_o <= scanline_s(to_integer(pixcnt_out))(1 downto 0);
			 hsync_o <= not scanline_s(to_integer(pixcnt_out))(8);
			 vsync_o <= not vsync_s;
			 pixcnt_out := pixcnt_out + 1;
			 if pixcnt_out=384 then
				pixcnt_out := (others => '0');
			 end if;
		  end if;
		end if;
		hsync := hsync_s;
	  end if;
	end process line_in;

end block scandoubler;
-------------------------------------------------------------------

end RTL;
